from .chassis import chassis
from .routing_engines import routing_engines
from .personality import personality
from .swver import software_version
from .ifd_style import ifd_style
from .switch_style import switch_style

FACT_LIST = [
  chassis,
  routing_engines,
  personality,
  software_version,
  ifd_style,
  switch_style,
]

__all__ = ['FACT_LIST']


# stdlib
from copy import deepcopy

# module libs
from .view import RunstatView
from .rsmfields import RunstatMakerViewFields
from .table import RunstatTable

class RunstatMaker(object):
  """
  Metaclass builder for RunstatTable and RunstatView.  The RunstatMaker is a namespace
  class used to build the following items:

    GetTable
      creates a "toplevel" class object for "getting" Junos run-state information,
      e.g. get-interface-information, get-route-information, etc.

    Table
      creates a table definition, used when the run-state information has one or
      more tables within the response.

    View 
      create a table view definition. a view creates the mapping of user-defined
      field names (i.e. what *they* want) to the underlying Junos XML.  the view
      is the abstraction layer between Python and Junos/XML essentially.  the
      view is defined by a dictionary of fields.

    Fields 
      creates the field definitions that construct a view.
  """

  @classmethod
  def GetTable(cls, cmd, args=None, args_key=None, item=None, key=RunstatTable.NAME_XPATH, view=None, table_name=None ):
    if table_name is None: table_name = "RunstatGetTable." + cmd
    new_cls = type(table_name, (RunstatTable,), {} )
    new_cls.GET_RPC = cmd
    new_cls.GET_ARGS = args or {}
    if args_key is not None: new_cls.GET_KEY = args_key
    new_cls.ITER_XPATH = item
    new_cls.NAME_XPATH = key 
    new_cls.VIEW = view
    new_cls.__module__ = __name__
    return new_cls

  @classmethod
  def Table(cls, item, key=RunstatTable.NAME_XPATH, view=None, table_name=None):
    if table_name is None: table_name = 'RunstatTable.' + item
    new_cls = type(table_name, (RunstatTable,), {} )
    new_cls.ITER_XPATH = item
    new_cls.NAME_XPATH = key 
    new_cls.VIEW = view
    new_cls.__module__ = __name__
    return new_cls

  @classmethod
  def View(cls, fields, **kvargs):
    """
    :fields: 
      dictionary of fields, structure of which is ~internal~ and should
      not be defined explicitly. use the RunstatMaker.Fields() mechanism to 
      create theserather than hardcoding the dictionary structures; 
      since they might change over time.

    :kvargs:
      'view_name' to name the class.  this could be useful for debug
      or eventual callback mechanisms.

      'groups' is a dict of name/xpath assocaited to fields
      this technique would be used to extract fields from 
      node-set elements like port <if-device-flags>.

      'extends' names the base View class to extend.  using this
      technique you can add to existing defined Views.
    """

    view_name = kvargs.get('view_name','RunstatView')
    new_cls = type(view_name, (RunstatView,), {})

    if 'extends' in kvargs:
      base_cls = kvargs['extends']
      new_cls.FIELDS = deepcopy(base_cls.FIELDS)
      new_cls.FIELDS.update(fields)
      if 'groups' in kvargs:
        new_cls.GROUPS = deepcopy(base_cls.GROUPS)
        new_cls.GROUPS.update(kvargs['groups'])
    else:
      new_cls.FIELDS = fields
      new_cls.GROUPS = kvargs['groups'] if 'groups' in kvargs else None    

    new_cls.__module__ = __name__
    return new_cls

  @classmethod
  def Fields(cls):
    """ class method that wraps the object instance """
    return RunstatMakerViewFields()







from os.path import dirname,join
from .yaml import loadyaml

# import the definitions from the YAML file in this directory and
# make them part of this module, yo!

_yml_file = 'lldp.yml'
globals().update( loadyaml(join(dirname(__file__),_yml_file)) )



from .util import Util

# @@@ may remove these and have folks import the specific files directly, TBD

import warnings

warnings.warn("util imports [Config,FS,SCP] will be removed in 0.0.3", DeprecationWarning)

from .config import Config
from .fs import FS
from .scp import SCP



# stdlib
from inspect import isclass

# 3rd-party modules
from lxml import etree

class RunstatTable(object):
  """
  DOCUMENT-ME
  """
  ITER_XPATH = None
  NAME_XPATH = 'name'
  VIEW = None

  ### -------------------------------------------------------------------------
  ### CONSTRUCTOR
  ### -------------------------------------------------------------------------

  def __init__(self, ncdev, table_xml = None):
    self._ncdev = ncdev
    self._xml_got = table_xml
    self.view = self.VIEW

  ### -------------------------------------------------------------------------
  ### PROPERTIES
  ### -------------------------------------------------------------------------

  @property
  def D(self):
    """ shortcut to the Device instance """
    return self._ncdev

  @property
  def R(self):
    """ shortcut to the Device RPC metaexec instance """
    return self._ncdev.rpc

  @property
  def xml(self):
    """ 
    returns the XML data received from the device.  for tables within
    tables, this XML will be the parent XML element that contains the
    table contents.
    """
    return self._xml_got

  @property
  def got(self):
    """ returns the XML data recieved from the device """
    return self.xml

  @property
  def view(self):
    """ returns the current view assigned to this table """
    return self._view

  @view.setter
  def view(self, cls):
    """ assigns a new view to the table """
    if cls is None:
      self._view = None
      return

    if not isclass(cls):
      raise ValueError("Must be given RunstatView class")

    self._view = cls

  @property 
  def is_container(self):
    """ 
    True if this table does not have records, but is a container of fields
    False otherwise
    """
    return self.ITER_XPATH is None

  @property 
  def can_refresh(self):
    """
    True if this table has an 'arg_key' that allows for simple retreval
    of items based on names
    """
    return hasattr(self,'GET_KEY')

  @property
  def _iter_xpath(self):
    """ internal use only """
    return self.got.xpath(self.ITER_XPATH) if self.got is not None else []
  
  ### -------------------------------------------------------------------------
  ### METHODS
  ### -------------------------------------------------------------------------

  def assert_data(self):
    """ ensure that the table has XML data """
    if self._xml_got is None: raise RuntimeError("No data")

  def _rpc_get(self, argkey=None, **kvargs):
    args = {}                     # create empty <dict>
    args.update(self.GET_ARGS)    # copy default args
    args.update(kvargs)           # copy caller provided args

    if hasattr(self, 'GET_KEY') and argkey is not None:
      args.update({self.GET_KEY: argkey })

    return getattr(self.R,self.GET_RPC)(**args)

  def get(self, *vargs, **kvargs):
    """ 
    Retrieve the XML table data from the Device instance and
    returns back the Table instance - for call-chaining purposes.  

    ALIAS: read, __call__    

    :vargs:
      [0] is the table :arg_key: value.  This is used so that
      the caller can retrieve just one item from the table without
      having to know the Junos RPC argument.

    :kvargs:
      these are the name/value pairs relating to the specific Junos
      XML command attached to the table.  For example, if the RPC 
      is 'get-route-information', there are parameters such as
      'table' and 'destination'.  Any valid RPC argument can be
      passed to :kvargs: to further filter the results of the :get():
      operation.  neato!

    NOTES:
      If you need to create a 'stub' for unit-testing
      purposes, you want to create a subclass of your table and 
      overload this methods.
    """

    argkey = vargs[0] if len(vargs) else None

    # execute the Junos RPC to retrieve the table
    self._xml_got = self._rpc_get(argkey, **kvargs)

    # returning self for call-chaining purposes, yo!
    return self

  def keys(self):
    """ returns a list of table item names """
    if self.is_container: return []
    return [n.findtext(self.NAME_XPATH).strip() for n in self._iter_xpath]

  def values(self):
    """ 
    returns list of table entry items().  
    """
    if self.view is None:
      # no View, so provide XML for each item
      return [this for this in self]
    else:
      # view object for each item
      return [this.items() for this in self]

  def items(self):
    """ returns list of tuple(name,values) for each table entry """
    return zip(self.keys(), self.values())

  ### -------------------------------------------------------------------------
  ### ALIASES
  ### -------------------------------------------------------------------------

  read = get

  ### -------------------------------------------------------------------------
  ### OVERLOADS
  ### -------------------------------------------------------------------------

  def __repr__(self):
    cname = self.__class__.__name__
    if self.is_container is not None:
      return "%s:%s: %s items" % (cname, self.D.hostname, len(self))
    else:
      return "%s:%s: data=%s" % (cname, self.D.hostname, ('no','yes')[self.got is not None])

  # make callable alias to :get(): 
  __call__ = get

  # ---------------------------------------------------------------------------
  # len is the number of items in the table
  # ---------------------------------------------------------------------------

  def __len__(self):
    return 1 if self.is_container else len(self._iter_xpath)

  # ---------------------------------------------------------------------------
  # [<name>]: select a table item based on <name>
  # [0]: gives the only view, for composite table
  # ---------------------------------------------------------------------------

  def __getitem__(self,value):
    """
    returns a table item.  if a table view is set (should be by default) then
    the item will be converted to the view upon return.  if there is no table 
    view, then the XML object will be returned.

    :value:
      when it is a string, this will perform a select based on the name
      when it is a number, this will perform a select based by position.
        nubers can be either positive or negative.
        [0] is the first item (first xpath is actually 1)
        [-1] is the last item
    """
    self.assert_data()

    use_view = self.view

    def get_xpath(find_value):
      if isinstance(find_value,str):
        # find by name
        xpath = self.ITER_XPATH + '[normalize-space(%s)="' % self.NAME_XPATH + find_value + '"]'
      elif isinstance(find_value,int):
        # find by index, assuming caller is using 0-index, and might use
        # negative values to reference from end of list
        xpath_pos = find_value + 1
        if find_value < 0:
          xpath_pos = len(self) + xpath_pos
        xpath = '%s[%s]' % (self.ITER_XPATH, xpath_pos)
      return xpath

    if self.ITER_XPATH is None:
      # this is a table of tables; i.e. not table of record views
      found = self.got
    else:
      if isinstance(value,tuple):
        # tuple(name,view_cls)
        use_view = value[1]        
        value = value[0]

      xpath = get_xpath(value)
      found = self.got.xpath(xpath)
      if not len(found): return None
      found = found[0]

    return use_view(table=self, view_xml=found) if use_view is not None else found

  # ---------------------------------------------------------------------------
  # iterate though each item in the talbe, not applicable for composite table
  # ---------------------------------------------------------------------------

  def __iter__(self):
    """ iteratable of each toplevel xpath item """
    self.assert_data()
    itself = lambda x: x
    view = lambda x: self.view(self, x)

    item_as = view if self.view else itself

    for this in self.got.xpath(self.ITER_XPATH):
      yield item_as(this)



import os
import sys

sys.path.insert(0,'lib')
from setuptools import setup, find_packages

setup(
    name = "junos-eznc",
    namespace_packages = ['jnpr'],
    version = "0.0.2",
    author = "Jeremy Schulman",
    author_email = "jschulman@juniper.net",
    description = ( "Junos 'EZ' automation for non-programmers" ),
    license = "BSD-2",
    keywords = "Junos NETCONF networking automation",
    url = "http://www.github.com/jeremyschulman/py-junos-eznc",
    package_dir={'':'lib'},    
    packages=find_packages('lib'),
    package_data={'jnpr.junos.op': ['*.yml']},
    install_requires=[
        "netaddr",
        "lxml",
        "jinja2",
        "scp"
    ],
)



# stdlib
import hashlib
from os import path

# 3rd-party modules
from lxml.builder import E

# local modules
from .util import Util
from .scp import SCP 

class SW(Util):
  """
  Softawre Utility class, used to perform a software upgrade and associated functions.

  Primary methods:
    sha256 - (class method) to compute SHA-256 hexdigest
    install - perform the entire software installation process
    reboot - reboots the system for the new image to take effect    

  Helpers called from install, but you can use these individually if needed:
    put - SCP put package file onto Junos device
    pkgadd - performs the 'request' operation to install the package
    validate - performs the 'request' to validate the package

  Other utils:
    rollback - same as 'request softare rollback'
    inventory - (property) provides file info for current and rollback 
                images on the device
  """

  @classmethod
  def sha256(cls,package):
    """
    computes the SHA-256 value on the package file.  This value should then be
    passed to the :install(): method.  This method would generally be used if 
    the same file was being loaded on a number of devices; thereby not re-calculating 
    the sha256 hexdigest each time the file was loaded onto a different box.

    :package:
      complete path to the package (*.tgz) file on the local server      
    """
    def hashfile(afile, hasher,blocksize=65536):
      buf = afile.read(blocksize)
      while len(buf) > 0:
          hasher.update(buf)
          buf = afile.read(blocksize)
      return hasher.hexdigest()      

    return hashfile(open(package,'rb'),hashlib.sha256())

  @classmethod
  def progress(cls,dev,report):
    """ simple progress report function """
    print dev.hostname + ": " + report

  ### -------------------------------------------------------------------------
  ### put - SCP put the image onto the device
  ### -------------------------------------------------------------------------

  def put(self, package, remote_path='/var/tmp', progress=None):
    """
    SCP 'put' the package file from the local server to the remote device.

    :remote_path:
      the directory on the device where the package will be copied to

    :progress:
      callback function to indicate progress.  see :install(): for details
    """
    def _progress(report): 
      if progress is not None: progress(self._dev, report)

    def _scp_progress(_path, _total, _xfrd):
      # init static variable
      if not hasattr(_scp_progress,'by10pct'): _scp_progress.by10pct = 0

      # calculate current percentage xferd
      pct = int(float(_xfrd)/float(_total) * 100)

      # if 10% more has been copied, then print a message
      if 0 == (pct % 10) and pct != _scp_progress.by10pct:
        _scp_progress.by10pct = pct
        _progress("%s: %s / %s (%s%%)" % (_path,_xfrd,_total,str(pct)))

    with SCP(self._dev, progress=_scp_progress) as scp:
      scp.put(package, remote_path)

  ### -------------------------------------------------------------------------
  ### pkgadd - used to perform the 'request system software add ...'
  ### -------------------------------------------------------------------------

  def pkgadd(self, remote_package ):    
    """ issue the 'request system software add' command on the package """

    rsp = self.rpc.request_package_add( no_validate=True, 
      package_name=remote_package )

    got = rsp.getparent()
    rc = int(got.findtext('package-result').strip())
    return True if rc == 0 else got.findtext('output').strip()

  ### -------------------------------------------------------------------------
  ### validate - perform 'request' operation to validate the package
  ### -------------------------------------------------------------------------

  def validate(self, remote_package):
    """ issues the 'request' operation to validate the package against the config """
    rsp = self.rpc.request_package_validate(package_name=remote_package).getparent()
    errcode = int(rsp.findtext('package-result'))
    return True if 0 == errcode else rsp.findtext('output').strip()

  ### -------------------------------------------------------------------------
  ### install - complete installation process, but not reboot
  ### -------------------------------------------------------------------------

  def install(self, package, remote_path='/var/tmp', validate=False, sha256=None, cleanfs=True, progress=None):
    """
    Performs the complete installation of the :package: that includes the following steps:
      (1) computes the SHA-256 checksum if not provided in :sha256:
      (2) performs a storage cleanup if :cleanfs: is True
      (3) SCP copies the package to the :remote_path: directory
      (4) validates the package if :validate: is True
      (5) installs the package

    You can get a progress report on this process by providing a :progress: callback;
    see description below.

    You will need to invoke the :reboot(): method explicity to reboot the device.

    :package: 
      is the install package tarball on the local filesystem.
    
    :remote_path: 
      is the directory on the Junos device where the package file will be SCP'd to.  
    
    :validate:
      determines whether or not to perform a config validation against the new image

    :sha256:
      SHA-256 hexdigest of the package file.  If this is not provided, then this
      method will perform the calculation.  If you are planning on using the
      same image for multiple updates, you should consider using the :sha256():
      method to precalculate this value and then provide to this method.

    :cleanfs:
      determines whether or not to perform a 'storeage cleanup' before SCP'ing 
      the file to the device.

    :progress:
      if provided, this is a callback function with a function prototype given
      the Device instance and the report string, e.g.
        
        def myprogress(dev, report):
          print "host: %s, report: %s" % (dev.hostname, report)
    """
    def _progress(report): 
      if progress is not None: progress(self._dev, report)
    
    rpc = self.rpc

    if sha256 is None: 
      _progress('computing local SHA-256 on: %s' % package)
      sha256 = SW.sha256(package)

    if cleanfs is True:
      _progress('cleaning filesystem ...')
      rpc.request_system_storage_cleanup()

    self.put( package, remote_path, progress)

    # validate checksum:
    remote_package = remote_path + '/' + path.basename(package)
    _progress('computing remote SHA-256 on: %s' % remote_package)
    rsp = rpc.get_sha256_checksum_information(path=remote_package)
    remote_sha256 = rsp.findtext('.//checksum').strip()

    if remote_sha256 != sha256:
      _progress("SHA-256 check failed.")
      return False
    _progress("SHA-256 check passed.")

    if validate is True:
      _progress("validating software against current config, please be patient ...")
      v_ok = self.validate(remote_package)
      if v_ok is not True:
        return v_ok # will be the string of output

    _progress("installing software ... this could take some time, please be patient ...")
    rsp = self.pkgadd( remote_package )
    return rsp

  ### -------------------------------------------------------------------------
  ### rebbot - system reboot
  ###
  ### @@@ - need to broader support to multi-RE devices
  ### -------------------------------------------------------------------------

  def reboot(self, in_min=0):    
    """ perform a system reboot, with optional delay (in minutes) """
    rsp = self.rpc(E('request-reboot', E('in', str(in_min))))
    return rsp.findtext('request-reboot-status').strip()

  ### -------------------------------------------------------------------------
  ### rollback - clears the install request
  ### -------------------------------------------------------------------------

  def rollback(self):
    """ 
    issues the 'request' command to do the rollback and returns the string
    output of the results
    """
    rsp = self.rpc.request_package_rollback()
    return rsp.text.strip()

  ### -------------------------------------------------------------------------
  ### inventory - file info on current and rollback packages
  ### -------------------------------------------------------------------------

  @property
  def inventory(self):
    """
    returns a dictionary of file listing information for current and rollback
    Junos install packages.  This information comes from the /packages directory.
    """
    from .fs import FS 
    fs = FS(self.dev)
    pkgs = fs.ls('/packages') 
    return dict(current=pkgs['files'].get('junos'), rollback=pkgs['files'].get('junos.old'))
